#!/usr/bin/env python3
"""
Python implementation (line-by-line) of TetrisGYM SPSv3 from v5.
by Javantea
July 24-27, 2022

Based on TetrisGYM/main.asm
Used Easy6502 to emulate the 6502 to understand the carry and get the correct
values for each intermediate. tetris_sps2.asm assembles in Easy6502 for your
use.

spawnID     := $0019
set_seed    := $0034 ; 3 bytes - rng_seed, rng_seed+1, spawnCount
set_seed_input := $0037 ; copied to set_seed during gameModeState_initGameState
tmp1        := $0000
tmp2        := $0001
tmp3        := $0002

030405 --> ijijljtislos
123456 --> ztzijijoztsosz
96c9df --> sojsllslztoilltzitioltztzo
a95292 --> zijloizioijl
7381fb --> lzlolzjisi
707c85 --> ijslzolzizso
4a96f0 --> lolositzlzoi
573fa4 --> zizltozzizsj
2ffdc5 --> sotltzjzjo
 -->
 -->
"""
import collections

class MemorySystem:
    def __init__(self, size):
        self.value = [0] * size
        self.offset = 0

    def getOffset(self, offset):
        v = MemorySystem(len(self.value))
        v.value = self.value
        v.offset = self.offset + offset
        return v

    def __getitem__(self, x):
        return self.value[x + self.offset]
    def __setitem__(self, x, value):
        self.value[x + self.offset] = value

#print('wtaf', ['{0:02x}'.format(r) for r in tmp1.value[0x34:0x3b]])

def ror(x, n):
    """
    Implementation of a valid (not-6502-compatible) 8-bit ror implementation.
    See the code that calls ror for more information on how it works.
    """
    mask = (1 << n) - 1
    return (x >> n) | ((x & mask) << (8-n))

class ThreadSafeSPS:
    def __init__(self, seed):
        self.seed = seed

        # While MemorySystem is not necessary, it allows to use the exact same variable names.
        # tmp1 is offset 0 (zero page)
        # tmp2 is offset 1 (address 0x01)
        # tmp3 is offset 2 (address 0x02)
        self.tmp1 = MemorySystem(256)
        self.tmp2 = self.tmp1.getOffset(1)
        self.tmp3 = self.tmp1.getOffset(2)
        # set_seed is offset 0x34 (address 0x34)
        self.set_seed = self.tmp1.getOffset(0x34)
        # set_seed_input is offset 0x37 (address 0x37)
        self.set_seed_input = self.tmp1.getOffset(0x37)
        self.change_seed(seed)
        # spawnID is at offset 0x19 (address 0x19)
        self.spawnID = self.tmp1.getOffset(0x19)

    def pickTetriminoSeed(self):
        """
        Python implementation (line-by-line) of TetrisGYM SPSv3 from v5.
        """
        self.setSeedNextRNG()

        #; SPSv3

        a = self.set_seed_input[2]
        # ror; ror; ror; ror
        a = ror(a, 4)
        #and #$F
        a &= 0xf
        #; v3
        #cmp #0
        #bne @notZero
        #lda #$10
        if False:
            a = 0x10
    #@notZero:
        #; v2
        #; cmp #0
        #; beq @compatMode

        #adc #1
        a += 1 + 1 # no idea why adding 1 adds 2 except carry.. When does carry get set?
        #sta tmp3 ; step + 1 in tmp3
        self.tmp3 = a
        #print("tmp3 0", tmp3)
    #@loop:
        while True:
            #jsr setSeedNextRNG
            self.setSeedNextRNG()
            #dec tmp3
            self.tmp3 -= 1
            #lda tmp3
            a = self.tmp3
            #bne @loop
            #print('loop', ['{0:02x}'.format(r) for r in tmp1.value[0x34:0x3b]])
            if a == 0: break

    #@compatMode:

        #inc set_seed+2 ; 'spawnCount'
        self.set_seed[2] += 1
        self.set_seed[2] &= 0xff
        #lda set_seed
        a = self.set_seed[0]
        #clc
        carry = 0
        #adc set_seed+2
        a += self.set_seed[2]
        #and #$07
        a &= 7
        #print('a 0', a)
        #cmp #$07
        #beq @invalidIndex
        if a != 7:
            #tax
            x = a
            #lda spawnTable,x
            a = spawnTable[x]
            #cmp spawnID
            #bne @useNewSpawnID
            if a != self.spawnID[0]:
                self.spawnID[0] = a
                return a
    #@invalidIndex:
        #ldx #set_seed
        x = 0x34
        #ldy #$02
        y = 2
        #jsr generateNextPseudorandomNumber
        self.generateNextPseudorandomNumber(x, y)
        #lda set_seed
        a = self.set_seed[0]
        #and #$07
        a &= 7
        #clc
        carry = 0
        #adc spawnID
        a += self.spawnID[0]
        a &= 0xff
    #@L992A:
        #cmp #$07
        #bcc @L9934
        while a >= 7:
            #sec FIXME:
            carry = 1
            #sbc #$07
            a -= 7
            a &= 0xff
            #jmp @L992A

    #@L9934:
        #tax
        x = a
        #lda spawnTable,x
        a = spawnTable[x]
    #@useNewSpawnID:
        #sta spawnID
        self.spawnID[0] = a
        #rts
        #print(tmp1.value)
        return a

    def setSeedNextRNG(self):
        """
        Call generateNextPseudorandomNumber(0x34, 2)
        """
        #ldx #set_seed
        x = 0x34
        #ldy #$02
        y = 2
        #jsr generateNextPseudorandomNumber
        self.generateNextPseudorandomNumber(x, y)
        #rts
        return

    #;reg x: zeropage addr of seed; reg y: size of seed
    def generateNextPseudorandomNumber(self, x, y):
        """
        Output
        X is the offset in memory of the seed. Usually 0x34
        y is the length of the seed. Usually 2.
        """
        #lda tmp1,x
        a = self.tmp1[x] # read the first byte of the seed
        #and #$02
        a &= 2 # reduce to a single bit
        #sta tmp1
        self.tmp1[0] = a # store in 0
        #lda tmp2,x
        a = self.tmp2[x] # read the second byte of the seed
        #and #$02
        a &= 2 # reduce to a single bit
        #eor tmp1
        a ^= self.tmp1[0] # xor the two bits together.
        #clc
        carry = 0
        #beq @updateNextByteInSeed
        if a != 0:
            #sec
            carry = 1
            #print('sec')
    #@updateNextByteInSeed:
        while True:
            #ror tmp1,x
            should_carry = self.tmp1[x] & 1
            self.tmp1[x] = ror(self.tmp1[x], 1)
            if carry:
                self.tmp1[x] |= 0x80 # FIXME: find the whole deal with this
            else:
                self.tmp1[x] &= 0x7f
            #print('ror', hex(x), hex(tmp1[x]))
            #print('debug', x, 'becomes', tmp1[x], tmp1.value[0x34:0x3b])
            carry = should_carry
            #print('sc', should_carry)
            # after ror tmp1,x we get 89 34 56 12 34 56 which explains why this RNG feels like nonsense.
            #inx
            x += 1
            #dey
            y -= 1
            #bne @updateNextByteInSeed
            if y == 0: break
        #print('eol')
        #rts
        return

    def change_seed(self, seed):
        """
        Modify the seed which is a global variable.
        """
        self.set_seed[0] = seed >> 16
        self.set_seed[1] = (seed >> 8) & 0xff
        self.set_seed[2] = seed & 0xff
        self.set_seed_input[0] = seed >> 16
        self.set_seed_input[1] = (seed >> 8) & 0xff
        self.set_seed_input[2] = seed & 0xff

#spawnTable
#        .byte   $02,$07,$08,$0A,$0B,$0E,$12
#        .byte   $02
spawnTable = [2,7,8,10,11,14,18,2]
# T J Z O S L I T

# ztzijijoztsosz

lookup = {8: 'Z', 2: 'T', 18: 'I', 7: 'J', 10: 'O', 11: 'S', 14: 'L'}

def main():
    import sys
    num_pieces = 14
    seed = 0x123456
    test_mode = False
    statistics = False
    if '-t' in sys.argv:
        sys.argv.remove('-t')
        test_mode = True
    if '-s' in sys.argv:
        sys.argv.remove('-s')
        statistics = True
    if len(sys.argv) > 1:
        seed = int(sys.argv[1], 16)
    if len(sys.argv) > 2:
        num_pieces = int(sys.argv[2])
    sps = ThreadSafeSPS(seed)

    if test_mode:
        orig_seed = (sps.set_seed[0], sps.set_seed[1], sps.set_seed[2])
        period_est = 0
        seen_before = set()
        seen_before.add(orig_seed)
        seen_before_l = [orig_seed]
        countdown = None
        first_repeating_val = None
        while True:
            piece = sps.pickTetriminoSeed()
            if countdown is None: period_est += 1
            seed_i = (sps.set_seed[0], sps.set_seed[1], sps.set_seed[2])
            if countdown is None and seed_i in seen_before:
                first_repeating_val = seed_i
                #print(seed_i)
                #break
                countdown = period_est + 100
            seen_before.add(seed_i)
            seen_before_l.append(seed_i)
            if period_est & 0xffff == 0:
                print(hex(period_est), seed_i, orig_seed)
            if countdown is not None:
                countdown -= 1
                if countdown == 0: break
        fout = open('test_out.txt','w')
        for i in seen_before_l:
            fout.write('{0!r}\n'.format(i))
        fout.close()
        first_pos = seen_before_l.index(first_repeating_val)
        period = seen_before_l.index(first_repeating_val, first_pos + 1) - first_pos
        print('Period:', seen_before_l.index(first_repeating_val, first_pos + 1), '-', first_pos, '=', period)
        full = seen_before_l[first_pos:first_pos+period]
        rep = seen_before_l[first_pos+period:first_pos+period+period]
        if full != rep:
            print("not same", len(full), len(rep))
        return

    v = ''.join([lookup[sps.pickTetriminoSeed()] for i in range(num_pieces)])
    print(v)
    if statistics:
        counter = collections.Counter()
        for piece in v:
            counter[piece] += 1
        print('T', '{0:03d}'.format(counter['T']))
        print('J', '{0:03d}'.format(counter['J']))
        print('Z', '{0:03d}'.format(counter['Z']))
        print('O', '{0:03d}'.format(counter['O']))
        print('S', '{0:03d}'.format(counter['S']))
        print('L', '{0:03d}'.format(counter['L']))
        print('I', '{0:03d}'.format(counter['I']))

if __name__ == '__main__':
    main()

